/*
 * firmware.c
 *
 * Created: 21.01.2023 21:58:18
 * Author : dUkk

output
AVR   74HC595
PA4 = SRCLK (CLOCK)
PB2 = RCLK
PA6 = SER (MOSI)
	
input
AVR   74HCT165
PA4 = CP (CLOCK)
PA7 = PL
PA5 = Q (MISO)
 */ 
#define F_CPU 1000000 //1Mhz

#include <avr/io.h>
#include <util/delay.h>

#define ACMAINS_PROPAGATION_DELAY 15
#define RELAY_TURNON_DELAY 20
#define RELAY_TURNOFF_DELAY 20
#define BISTABLE_RELAY_TURN_DELAY 20

//74HCT595 output
unsigned char outstate = 0, instate = 0;

void readwritebyte()
{
	//pulse parallel load to latch state of inputs HC165
	PORTA &= ~(1<<PORTA7);
	PORTA |= (1<<PORTA7);

	//latch off HC595
	PORTB &= ~(1<<PORTB2);

	//to bus
	USIDR = outstate;
	USISR |= (1<<USIOIF); // clear counter and counter overflow interrupt flag
	while ( !(USISR & (1<<USIOIF)) ) USICR |= (1<<USITC);
	//from bus
	instate = USIDR;
	
	//latch on HC595
	PORTB |= (1<<PORTB2);
}

void emergencyOFF()
{
	//first turn off ALL phases in hybrid relays
	outstate = 0;
	readwritebyte();
	_delay_ms(RELAY_TURNOFF_DELAY);
	//next shut off one by one power bistable relays
	PORTA |= (1<<PORTA3);
	_delay_ms(BISTABLE_RELAY_TURN_DELAY);
	PORTA &= ~(1<<PORTA3);
	PORTA |= (1<<PORTA2);
	_delay_ms(BISTABLE_RELAY_TURN_DELAY);
	PORTA &= ~(1<<PORTA2);
	PORTA |= (1<<PORTA1);
	_delay_ms(BISTABLE_RELAY_TURN_DELAY);
	PORTA &= ~(1<<PORTA1);
	//here we already should be shutted down because PA1 connected relay powered us
	while(1) {};
}

int main(void)
{
	//disable unused peripherals
	PRR = (1<<PRADC) | (1<<PRTIM1);// | (1<<PRTIM0);
	//disable all pullups
	MCUCR |= (1<<PUD);
	//setup OUTputs on portB PB2 and portA (PA1,PA2,PA3, PA4,PA5,PA7)
	DDRA = (1<<DDA1) | (1<<DDA2) | (1<<DDA3) | (1<<DDA4) | (1<<DDA5) | (1<<DDA7);
	//set DI pin as INput
	DDRA &= ~(1<<DDA6);    
	//PORTA |= (1<<PORTA6); //pull-up
	DDRB = (1<<DDB2);
	
	//setup USI SPI
	USICR &= ~( (1<<USISIE) | (1<<USIOIE) | (1<<USIWM1) | (1<<USICS0)); //probably not needed
	USICR |= (1<<USIWM0) | (1<<USICS1) | (1<<USICLK);  //3wire mode, clock source is 4bit counter

    while (1) 
    {
		//write output lines byte and read input lines byte
		readwritebyte();

		//D0 - heater1 requested state ON
		if(instate & (1<<1)) {	
			//when it was OFF -> transition to ON state
			if((outstate & (1<<7)) == 0) {
				//triac on  QG
				outstate |= (1<<6);
				readwritebyte();
				_delay_ms(ACMAINS_PROPAGATION_DELAY);
				//relay on  QH
				outstate |= (1<<7);
				readwritebyte();
				_delay_ms(RELAY_TURNON_DELAY);
				//triac off QG
				outstate &= ~(1<<6);
				readwritebyte();
				_delay_ms(ACMAINS_PROPAGATION_DELAY*2);
				
				//readback cur state
				readwritebyte();
				//D6 should be LOW or if HIGH - fault
				if(instate & (1<<7)) emergencyOFF();
			}
		} else {
			//when it was ON -> transition to OFF state
			if(outstate & (1<<7)) {
				//triac on  QG
				outstate |= (1<<6);
				readwritebyte();
				_delay_ms(ACMAINS_PROPAGATION_DELAY);
				//relay off QH
				outstate &= ~(1<<7);
				readwritebyte();
				_delay_ms(RELAY_TURNOFF_DELAY);
				//triac off QG
				outstate &= ~(1<<6);
				readwritebyte();
				_delay_ms(ACMAINS_PROPAGATION_DELAY*2);
				
				//readback cur state
				readwritebyte();
				//D6 should be HIGH or if LOW - fault
				if((instate & (1<<7)) == 0) emergencyOFF();
			}
		}
		
		//D1 heater2 requested state ON
		if(instate & (1<<2)) {
			//when it was OFF -> transition to ON state
			if((outstate & (1<<5)) == 0) {
				//triac on  QE
				outstate |= (1<<4);
				readwritebyte();
				_delay_ms(ACMAINS_PROPAGATION_DELAY);
				//relay on  QF
				outstate |= (1<<5);
				readwritebyte();
				_delay_ms(RELAY_TURNON_DELAY);
				//triac off QE
				outstate &= ~(1<<4);
				readwritebyte();
				_delay_ms(ACMAINS_PROPAGATION_DELAY*2);
				
				//readback cur state
				readwritebyte();
				//D5 should be LOW or if HIGH - fault
				if(instate & (1<<6)) emergencyOFF();
			}			
		} else {
			//when it was ON -> transition to OFF state
			if(outstate & (1<<5)) {
				//triac on  QE
				outstate |= (1<<4);
				readwritebyte();
				_delay_ms(ACMAINS_PROPAGATION_DELAY);
				//relay off QF
				outstate &= ~(1<<5);
				readwritebyte();
				_delay_ms(RELAY_TURNOFF_DELAY);
				//triac off QE
				outstate &= ~(1<<4);
				readwritebyte();
				_delay_ms(ACMAINS_PROPAGATION_DELAY*2);
				
				//readback cur state
				readwritebyte();
				//D5 should be HIGH or if LOW - fault
				if(!(instate & (1<<6))) emergencyOFF();
			}
		}
		
		//D2 heater3 requested state ON
		if(instate & (1<<3)) {
			//D4 (4th bit of register) is HIGH if voltage NOT detected
			
			//when it was OFF -> transition to ON state
			if((outstate & (1<<3)) == 0) {
				//triac on  QC
				outstate |= (1<<2);
				readwritebyte();
				_delay_ms(ACMAINS_PROPAGATION_DELAY);
				//relay on  QD
				outstate |= (1<<3);
				readwritebyte();
				_delay_ms(RELAY_TURNON_DELAY);
				//triac off QC
				outstate &= ~(1<<2);
				readwritebyte();
				_delay_ms(ACMAINS_PROPAGATION_DELAY*2);
				
				//readback cur state
				readwritebyte();
				//D4 should be LOW or if HIGH - fault
				if(instate & (1<<5)) emergencyOFF();
			}
		} else {
			//when it was ON -> transition to OFF state
			if(outstate & (1<<3)) {
				//triac on  QC
				outstate |= (1<<2);
				readwritebyte();
				_delay_ms(ACMAINS_PROPAGATION_DELAY);
				//relay off QD
				outstate &= ~(1<<3);
				readwritebyte();
				_delay_ms(RELAY_TURNOFF_DELAY);
				//triac off QC
				outstate &= ~(1<<2);
				readwritebyte();
				_delay_ms(ACMAINS_PROPAGATION_DELAY*2);
				
				//readback cur state
				readwritebyte();
				//D4 should be HIGH or if LOW - fault
				if(!(instate & (1<<5))) emergencyOFF();
			}
		}
		
    }
}
